
import re
import functools

import mpd
from PySide.QtCore import *

import logging
logger = logging.getLogger('qygmy')


class State(QObject):

    default = None
    changed = None
    changed2 = None

    def __init__(self):
        super().__init__()
        self._value = self.normalize(self.default)

    @property
    def value(self):
        return self._value

    def update(self, new_value, force=False):
        new_value = self.normalize(new_value)
        if self._value != new_value or force:
            old_value = self._value
            self._value = new_value
            self.changed.emit(self._value)
            self.changed2.emit(self._value, old_value)

    def send(self, value):
        pass

    def normalize(self, value):
        if value is None or value == '':
            value = self.default
        if isinstance(value, str):
            value = value.strip()
        if not isinstance(value, type(self.default)):
            value = type(self.default)(value)
        return value

    def prepare(self, value):
        return self.normalize(value)

    @classmethod
    def create(cls, parent, name, init, send_cmd=None):
        if not callable(send_cmd):
            send_cmd = lambda v, s=send_cmd or name: getattr(parent.conn, s)(v)
        class CompleteState(cls):
            default = init
            changed = Signal(type(init))
            changed2 = Signal(type(init), type(init))
            def send(self, value):
                value = self.prepare(value)
                mpd_cmd(lambda p, v: send_cmd(v))(parent, value)
        s = CompleteState()
        s.setObjectName(name)
        s.setParent(parent)
        setattr(parent, name, s)


class BoolState(State):
    def normalize(self, value):
        return value in (True, 1, '1')

    def prepare(self, value):
        return int(self.normalize(value))

    @classmethod
    def create(cls, parent, name, init=False, send_cmd=None):
        super().create(parent, name, init, send_cmd)


class TimeTupleState(State):
    def normalize(self, value):
        if isinstance(value, (tuple, list)):
            v = [0 if i in ('', None, -1, '-1') else int(i) for i in value]
            v = v[:2]
            v = v + [0]*(2 - len(v))
            return v
        return self.default

    def prepare(self, value):
        return int(value)


class PathState(State):
    def normalize(self, value):
        v = super().normalize(value)
        return v.strip('/')


def mpd_cmd(method=None, ignore_conn=False, fallback=None):
    if method is None:
        def decorator(method):
            return mpd_cmd(method, ignore_conn, fallback)
        return decorator
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        ret = fallback
        if self.state.value != 'disconnect' or ignore_conn:
            try:
                ret = method(self, *args, **kwargs)
            except (mpd.MPDError, OSError) as e:
                self.handle_error(e)
        self.update_state()
        return ret
    return wrapper


def simple_mpd(name, *args):
    return mpd_cmd(lambda self: getattr(self.conn, name)(*args))


def not_callable(*args):
    raise TypeError("Not callable")


class MPDCommandList:
    def __init__(self, conn):
        self.conn = conn
    def __enter__(self):
        self.conn.command_list_ok_begin()
    def __exit__(self, *exc):
        self.conn.command_list_end()


class Connection:

    MPD_COMMAND_ERROR_RE = re.compile(r'^\[([0-9]+)@([0-9]+)\] \{([^}]*)\} (.*)$')
    MAX_MPD_ARGUMENTS = 36

    def parse_exception(self, exc):
        m = self.MPD_COMMAND_ERROR_RE.match(exc.args[0])
        if m is None:
            return '', '', '', str(exc)
        return m.groups()

    def handle_error(self, exc):
        if isinstance(exc, (mpd.ConnectionError, mpd.ProtocolError, OSError)):
            # Critical errors
            try:
                self.conn.disconnect()
            except Exception as e:
                pass
            # Needed because it isn't called when socket's close()
            # raises an exception (python-mpd2 bug(?)).
            self.conn._reset()
            self.state.update('disconnect')
        logger.error('{}: {}'.format(exc.__class__.__name__, exc))
        msg = None
        if isinstance(exc, mpd.CommandError):
            errno, cmdno, cmd, msg = self.parse_exception(exc)
        elif isinstance(exc, OSError) and exc.errno is not None:
            msg = exc.strerror
        elif isinstance(exc, (mpd.ConnectionError, OSError)):
            msg = str(exc)
        if msg:
            self.main.error(msg[0].upper() + msg[1:] + ('.' if msg[-1] != '.' else ''))


class ProperConnection(Connection):
    def __init__(self, main):
        super().__init__(main)
        self.main = main
        self.conn = mpd.MPDClient()
        self.mpd_cmdlist = MPDCommandList(self.conn)
        State.create(self, 'state', 'disconnect', not_callable)

    def update_state(self):
        pass


class RelayingConnection(Connection):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent

    @property
    def main(self):
        return self.parent.main

    @property
    def conn(self):
        return self.parent.conn

    @property
    def mpd_cmdlist(self):
        return self.parent.mpd_cmdlist

    @property
    def state(self):
        return self.parent.state

    def update_state(self):
        self.parent.update_state()




#!/usr/bin/env python3

import re

from PySide.QtGui import QApplication

from . import functions


IDENTIFIER = r'([_a-zA-Z][_a-zA-Z0-9]*)'


class TemplateError(Exception):
    pass


class Renderable:
    def render(self, context):
        return ''


class Text(Renderable):
    def __init__(self, text):
        self.text = text

    def render(self, context):
        return self.text

    def __repr__(self):
        return repr(self.text)


class Element(Renderable):

    def __init__(self, src, start, end):
        self.src = src
        self.start = start
        self.end = end

    @classmethod
    def parse_first(cls, src, pos=0):
        return None

    def parse_rest(self):
        pass


class RegexElement(Element):

    regex = None

    @classmethod
    def parse_first(cls, src, pos=0):
        m = cls.regex.search(src, pos)
        if m is None:
            return None
        return cls(src, m.start(), m.end(), *m.groups())


class Backslash(RegexElement):

    regex = re.compile(r'\\(.)')

    def __init__(self, src, start, end, char):
        super().__init__(src, start, end)
        self.char = char

    def render(self, context):
        return self.char

    def __repr__(self):
        return 'Backslash({!r})'.format(self.char)


class Variable(RegexElement):

    regex = re.compile(r'%' + IDENTIFIER + r'%')

    def __init__(self, src, start, end, name):
        super().__init__(src, start, end)
        self.name = name.lower()

    def render(self, context):
        return str(context.get(self.name, ''))

    def __repr__(self):
        return 'Variable({!r})'.format(self.name)


class Sequence(Element):

    class End(RegexElement):
        regex = re.compile(r'$')

    def __init__(self, src, start, end, delims=[], types=None):
        super().__init__(src, start, end)
        if types is None:
            types = [Backslash, Variable, Function]
        self.delims = delims + [self.End]
        self.types = types + self.delims
        self.elems = []

    @classmethod
    def parse_first(cls, src, pos=0, delims=[], types=None):
        return cls(src, pos, pos, delims, types)

    def parse_rest(self):
        while True:
            matches = [t.parse_first(self.src, self.end) for t in self.types]
            minimum = None
            for m in matches:
                if m is not None and (minimum is None or m.start < minimum.start):
                    minimum = m
            if self.end < minimum.start:
                self.elems.append(Text(self.src[self.end:minimum.start]))
            minimum.parse_rest()
            self.end = minimum.end
            if type(minimum) in self.delims:
                self.last = minimum
                return
            self.elems.append(minimum)

    def render(self, context):
        return ''.join(e.render(context) for e in self.elems)

    def __repr__(self):
        return '+'.join(repr(e) for e in self.elems)


class Function(RegexElement):

    regex = re.compile(r'\$' + IDENTIFIER + r'\(')


    class Comma(RegexElement):
        regex = re.compile(r',')


    class ClosingParen(RegexElement):
        regex = re.compile(r'\)')


    def __init__(self, src, start, end, name):
        super().__init__(src, start, end)
        self.name = name.lower()
        self.args = []

        self._ctx_additions = {
            '__timefmt__': (
                QApplication.translate('templates', '{}{}d {:02d}:{:02d}:{:02d}'),
                QApplication.translate('templates', '{}{}:{:02d}:{:02d}'),
                QApplication.translate('templates', '{}{:02d}:{:02d}'),
            ),
        }

        f = getattr(functions, 'lazy_' + self.name, None)
        if f is not None:
            self.function = f
            self.call = self.call_lazy
            return
        f = getattr(functions, 'f_' + self.name, None)
        if f is not None:
            self.function = f
            self.call = self.call_f
            return
        f = getattr(functions, 'context_' + self.name, None)
        if f is not None:
            self.function = f
            self.call = self.call_context

    def parse_rest(self):
        while True:
            seq = Sequence.parse_first(self.src, self.end,
                    delims=[self.Comma, self.ClosingParen])
            seq.parse_rest()
            self.end = seq.end
            self.args.append(seq)
            if type(seq.last) != self.Comma:
                return

    def call_lazy(self, context):
        args = [(lambda a=a: a.render(context)) for a in self.args]
        return self.function(*args)

    def call_f(self, context):
        args = [a.render(context) for a in self.args]
        return self.function(*args)

    def call_context(self, context):
        context.update(self._ctx_additions)
        args = [a.render(context) for a in self.args]
        return self.function(context, *args)

    def render(self, context):
        try:
            return self.call(context)
        except Exception as e:
            raise TemplateError(str(e)) from e
        raise TemplateError('no such function: {!r}'.format(self.name))

    def __repr__(self):
        s = 'Function('
        s += ','.join(repr(i) for i in [self.name] + self.args)
        return s + ')'


class Template:
    def __init__(self, source):
        self.compiled = Sequence.parse_first(source)
        self.compiled.parse_rest()

    def render(self, context):
        r = []
        for e in self.compiled.elems:
            try:
                r.append(e.render(context))
            except TemplateError as e:
                r.append('##error: {}##'.format(str(e)))
        return ''.join(r)

    def __repr__(self):
        return 'Template({!r})'.format(self.compiled)





from PySide.QtCore import QObject
from PySide.QtGui import QApplication, QIcon, QBrush, QColor

from .templates import Template


class Formatter(QObject):

    def __init__(self, settings):
        super().__init__()
        self.settings = settings
        self._tmplcache = {}
        self.retranslate()

    _directory_icon = QIcon.fromTheme('folder')
    _playlist_icon = QIcon.fromTheme('text-plain')
    _high_prio_background = QBrush(QColor(255, 255, 0, 127))

    playlist_column_count = 2

    def retranslate(self):
        self.templates = {
            'current_song_tooltip': self.tr(
                '$if3(%paused%,[Paused] ,'
                    '%stopped%,[Stopped] ,'
                    '%connected%,[Connected] ,'
                    '%disconnected%,[Disconnected])'
                '$if2(%file%,%directory%,%playlist%)'
            ),
            'playlist_column2': self.tr('$time(%length%)'),
            'playlist_tooltip': self.tr(
                '$if($ne(%prio%,0),[High priority] ,)'
                '$if2(%file%,%directory%,%playlist%)'
            ),
            'status': self.tr(
                '$if(%disconnected%,,'
                    '%totalcount% songs'
                    '$if($gt(%totallength%,0),'
                        '\\, $time(%totallength%) total,))'
            ),
            'statistics': (
                (self.tr('Songs:'), self.tr('%songs%')),
                (self.tr('Albums:'), self.tr('%albums%')),
                (self.tr('Artists:'), self.tr('%artists%')),
                (self.tr('Uptime:'), self.tr('$time(%uptime%)')),
                (self.tr('DB playtime:'), self.tr('$time(%db_playtime%)')),
                (self.tr('This instance:'), self.tr('$time(%playtime%)')),
            ),
            'details': (
                (self.tr('%file%'), None),
                (self.tr('Title:'), self.tr('%title%')),
                (self.tr('Artist:'), self.tr('%artist%')),
                (self.tr('Album:'), self.tr('%album%')),
                (self.tr('Date:'), self.tr('%date%')),
                (self.tr('Track:'), self.tr('%track%$if(%totaltracks%, / %totaltracks%,)')),
                (self.tr('Disc:'), self.tr('%disc%$if(%totaldiscs%, / %totaldiscs%,)')),
                (self.tr('Comment:'), self.tr('%comment%')),
                (self.tr('Length:'), self.tr('$if(%length%,$time(%length%),)')),
                (self.tr('Last modified:'), self.tr('%lastmodified%')),
                (self.tr('Composer:'), self.tr('%composer%')),
                (self.tr('Performer:'), self.tr('%performer%')),
            )
        }

    standard_tags = {'file', 'directory', 'playlist', 'filename', 'title',
        'artist', 'album', 'date', 'track', 'totaltracks', 'disc', 'totaldiscs',
        'comment', 'length', 'lastmodified', 'composer', 'performer',
    }

    def _escape(self, s):
        return s.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')

    def _prepare_song(self, song, html=True):
        song = song.copy()
        song.pop('id', None)
        song.pop('pos', None)
        if 'prio' not in song:
            song['prio'] = '0'
        for key in ['file', 'directory', 'playlist']:
            if key in song:
                filename = song[key].rsplit('/', 1)[-1]
                if filename:
                    song['filename'] = filename
        if 'time' in song:
            song['length'] = int(song.pop('time'))
        if 'track' in song and '/' in song['track']:
            song['track'], song['totaltracks'] = song['track'].split('/', 1)
        if 'disc' in song and '/' in song['disc']:
            song['disc'], song['totaldiscs'] = song['disc'].split('/', 1)
        if 'last-modified' in song:
            song['lastmodified'] = song.pop('last-modified').replace('T', ' ').replace('Z', '')
        ret = {}
        for k, v in song.items():
            if isinstance(v, (list, tuple)):
                v = ', '.join(v)
            else:
                v = str(v)
            if html:
                v = self._escape(v)
            if v != '':
                ret[k] = v
        return ret

    def _prepare_state(self, state):
        return {{
            'play': 'playing',
            'pause': 'paused',
            'stop': 'stopped',
            'disconnect': 'disconnected',
            'connect': 'connected',
        }[state]: '1'}

    def _prepare_times(self, elapsed, total):
        return {
            'elapsed': str(elapsed),
            'total': str(total),
        }

    def _compile(self, template):
        pref = '$if(%bold%,<span style="font-weight: bold">,)'
        suff = '$if(%bold%,</span>,)'
        if isinstance(template, str):
            return Template(pref + template + suff)
        r = []
        for a, b in template:
            a = self._compile(a)
            if b is not None:
                b = self._compile(b)
            r.append((a, b))
        return tuple(r)

    def _render(self, template, context):
        if isinstance(template, Template):
            return template.render(context)
        r = []
        for a, b in template:
            a = self._render(a, context)
            if b is not None:
                b = self._render(b, context)
            r.append((a, b))
        return r

    def render(self, name, context, bold=False):
        tmpl = self._tmplcache.get(name)
        if tmpl is None:
            if name in self.settings['format']:
                t = self.settings['format'][name]
            else:
                t = self.templates[name]
            tmpl = self._tmplcache[name] = self._compile(t)
        context['bold'] = '1' if bold else ''
        return self._render(tmpl, context)

    def clear_cache(self):
        self._tmplcache.clear()

    def window_title(self, state, song):
        context = self._prepare_state(state)
        context.update(self._prepare_song(song, html=False))
        return self.render('window_title', context)

    def progressbar(self, state, elapsed, total, song):
        context = self._prepare_state(state)
        context.update(self._prepare_times(elapsed, total))
        context.update(self._prepare_song(song, html=False))
        return self.render('progressbar', context)

    def current_song(self, state, song):
        context = self._prepare_state(state)
        context.update(self._prepare_song(song))
        return self.render('current_song', context)

    def current_song_tooltip(self, state, song):
        context = self._prepare_state(state)
        context.update(self._prepare_song(song, html=False))
        return self.render('current_song_tooltip', context)

    def playlist_item(self, song, column, is_current):
        if column == 0:
            return self.render('playlist_item', self._prepare_song(song), bold=is_current)
        elif column == 1 and 'time' in song:
            return self.render('playlist_column2', {'length': song['time']}, bold=is_current)
        return ''

    def playlist_icon(self, song, column, is_current):
        if column == 0:
            if 'playlist' in song:
                return self._playlist_icon
            if 'directory' in song:
                return self._directory_icon

    def playlist_tooltip(self, song, column, is_current):
        return self.render('playlist_tooltip', self._prepare_song(song, html=False))

    def playlist_background(self, song, column, is_current):
        if song.get('prio', '0') != '0':
            return self._high_prio_background

    def status(self, state, totallength, totalcount):
        context = {'totallength': str(totallength), 'totalcount': str(totalcount)}
        context.update(self._prepare_state(state))
        return self.render('status', context)

    def info_dialog(self, name, info):
        unrecognized = []
        if name == 'details':
            info = self._prepare_song(info, html=False)
            info.pop('prio', None)
            for k, v in info.items():
                if k not in self.standard_tags and k and v:
                    unrecognized.append((k + ':', v))
        r = [i for i in self.render(name, info) if i[1] != '']
        return r + unrecognized





from PySide.QtCore import *
from PySide.QtGui import *

from .ui.browser import Ui_browser


class Browser(QMainWindow):
    def __init__(self, main):
        super().__init__()
        self.main = main
        self.srv = main.srv
        self.setup_ui()

    def setup_ui(self):
        self.ui = Ui_browser()
        self.ui.setupUi(self)
        self.setup_icons()
        self.ui.database.setup(self.srv.database)
        self.srv.database.current.changed.connect(self.ui.dbpath.set_path)
        self.ui.dbpath.clicked.connect(self.srv.database.cd)
        self.ui.playlists.setup(self.srv.playlists)
        self.srv.playlists.current.changed.connect(self.ui.plpath.set_path)
        self.ui.plpath.clicked.connect(self.srv.playlists.cd)
        self.ui.search_results.setup(self.srv.search)
        self.ui.search_button.setDefaultAction(self.ui.search)
        self.srv.state.changed.connect(self.on_state_changed)

        try:
            if 'browser_geometry' in self.main.settings['guistate']:
                self.restoreGeometry(QByteArray.fromBase64(self.main.settings['guistate']['browser_geometry']))
            if ('browser_tab' in self.main.settings['guistate'] and
                    0 <= int(self.main.settings['guistate']['browser_tab']) < self.ui.tabs.count()):
                self.ui.tabs.setCurrentIndex(int(self.main.settings['guistate']['browser_tab']))
        except:
            pass

    def setup_icons(self):
        self.setWindowIcon(QIcon.fromTheme('list-add'))
        for action, icon in (
            ('add', 'list-add'),
            ('remove', 'edit-delete'),
            ('close', 'window-close'),
            ('search', 'edit-find'),
        ):
            getattr(self.ui, action).setIcon(QIcon.fromTheme(icon))
        for i, icon in enumerate(['folder-sound', 'document-multiple', 'edit-find']):
            self.ui.tabs.setTabIcon(i, QIcon.fromTheme(icon))
        self.ui.dbpath.set_root_icon('folder-sound')
        self.ui.plpath.set_root_icon('document-multiple')

    @property
    def current_view(self):
        i = self.ui.tabs.currentIndex()
        if i == 0: return self.ui.database
        elif i == 1: return self.ui.playlists
        elif i == 2: return self.ui.search_results

    def closeEvent(self, e):
        self.main.settings['guistate']['browser_geometry'] = str(self.saveGeometry().toBase64())
        self.main.settings['guistate']['browser_tab'] = str(self.ui.tabs.currentIndex())
        super().closeEvent(e)

    def contextMenuEvent(self, e):
        e.accept()
        self.ui.add.setVisible(self.current_view.can_add_to_queue())
        self.ui.remove.setVisible(self.current_view.can_remove())
        self.ui.rename.setVisible(self.current_view.can_rename())
        self.ui.details.setVisible(self.current_view.details() is not None)
        self.ui.context_menu.popup(e.globalPos())

    def on_state_changed(self, state):
        c = state != 'disconnect'
        for act in ('search', 'add', 'addplay', 'replace', 'replaceplay',
                'remove', 'rename', 'details'):
            getattr(self.ui, act).setEnabled(c)
        self.ui.dbpath.setEnabled(c)
        self.ui.plpath.setEnabled(c)
        self.ui.what.setEnabled(c)
        self.ui.query.setEnabled(c)

    @Slot()
    def on_close_triggered(self):
        self.close()

    @Slot()
    def on_search_triggered(self):
        self.srv.search.cd((self.ui.what.currentIndex(), self.ui.query.text()))

    @Slot()
    def on_add_triggered(self, play=False, replace=False):
        self.current_view.add_selected_to_queue(play, replace)

    @Slot()
    def on_addplay_triggered(self):
        self.on_add_triggered(True)

    @Slot()
    def on_replace_triggered(self):
        self.on_add_triggered(False, True)

    @Slot()
    def on_replaceplay_triggered(self):
        self.on_add_triggered(True, True)

    @Slot()
    def on_remove_triggered(self):
        self.current_view.remove_selected()

    @Slot()
    def on_rename_triggered(self):
        self.current_view.rename_selected()

    @Slot()
    def on_details_triggered(self):
        d = self.current_view.details()
        if d is not None:
            self.main.info.exec_('details', d, self.tr('Details'))





from PySide.QtCore import *
from PySide.QtGui import *


class RichTextDelegate(QStyledItemDelegate):
    def _make_document(self, text):
        doc = QTextDocument()
        doc.setHtml(text)
        doc.setDocumentMargin(3)
        return doc

    def paint(self, painter, option, index):
        option = QStyleOptionViewItemV4(option)
        self.initStyleOption(option, index)
        style = QApplication.style() if option.widget is None else option.widget.style()
        doc = self._make_document(option.text)
        option.text = ''
        style.drawControl(QStyle.CE_ItemViewItem, option, painter)
        ctx = QAbstractTextDocumentLayout.PaintContext()
        if option.state & QStyle.State_Selected:
            ctx.palette.setColor(QPalette.Text, option.palette.color(QPalette.Active, QPalette.HighlightedText))
        textRect = style.subElementRect(QStyle.SE_ItemViewItemText, option)
        painter.save()
        painter.translate(textRect.topLeft())
        painter.setClipRect(textRect.translated(-textRect.topLeft()))
        doc.documentLayout().draw(painter, ctx)
        painter.restore()

    def sizeHint(self, option, index):
        option = QStyleOptionViewItemV4(option);
        self.initStyleOption(option, index);
        doc = self._make_document(option.text);
        return QSize(doc.size().width(), doc.size().height())


class ClickableProgressBar(QProgressBar):
    def mousePressEvent(self, e):
        sz = self.size()
        if e.button() == Qt.LeftButton and 0 <= e.x() < sz.width() and 0 <= e.y() < sz.height():
            e.accept()
            minv, maxv = self.minimum(), self.maximum()
            newvalue = minv + round((e.x() / sz.width()) * (maxv - minv))
            self.clicked.emit(newvalue)
        else:
            e.ignore()

    clicked = Signal(int)


class ChangeDuringDragTabBar(QTabBar):
    def __init__(self, parent=None, drag_change_interval=1000):
        super().__init__(parent)
        self.setAcceptDrops(True)
        self._dragtimer = QTimer(self)
        self._dragtimer.setInterval(drag_change_interval)
        self._dragtimer.setSingleShot(True)
        self._dragtab = -1
        self._dragtimer.timeout.connect(self._change_tab)

    def _change_tab(self):
        if self._dragtab >= 0:
            self.setCurrentIndex(self._dragtab)

    def dropEvent(self, e):
        self._dragtimer.stop()
        self._dragtab = -1
        super().dropEvent(e)

    def dragLeaveEvent(self, e):
        self._dragtimer.stop()
        self._dragtab = -1
        super().dragLeaveEvent(e)

    def dragEnterEvent(self, e):
        self._dragtimer.stop()
        self._dragtab = self.tabAt(e.pos())
        if self._dragtab >= 0:
            self._dragtimer.start()
        e.accept()
        super().dragEnterEvent(e)

    def dragMoveEvent(self, e):
        t = self.tabAt(e.pos())
        if t != self._dragtab:
            self._dragtimer.stop()
            self._dragtab = t
            if t >= 0:
                self._dragtimer.start()
        super().dragMoveEvent(e)


class ChangeDuringDragTabWidget(QTabWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._tabbar = ChangeDuringDragTabBar(self)
        self.setTabBar(self._tabbar)


class SonglistView(QTreeView):

    selection_changed = Signal()

    def __init__(self, parent):
        super().__init__(parent)
        self._enter_action = QAction('Add or descend into', self)
        self._enter_action.setShortcut(QKeySequence(self.tr('Return')))
        self._enter_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.addAction(self._enter_action)

    def setup(self, model):
        self.setModel(model)
        self._delegate = RichTextDelegate()
        self.setItemDelegate(self._delegate)

        h = self.header()
        h.setResizeMode(0, QHeaderView.Stretch)
        for i in range(1,h.count()):
            h.setResizeMode(i, QHeaderView.ResizeToContents)

        self.doubleClicked.connect(self._double_clicked)
        self._enter_action.triggered.connect(self._enter_pressed)

    def selectionChanged(self, selected, deselected):
        self.selection_changed.emit()
        super().selectionChanged(selected, deselected)

    def selection(self):
        return sorted(i.row()
                for i in self.selectedIndexes()
                if i.column() == 0)

    def _double_clicked(self, index):
        self.model().item_chosen(index.row())

    def _enter_pressed(self):
        if self.state() == self.EditingState:
            index = self.currentIndex()
            editor = self.indexWidget(index)
            if editor:
                self.commitData(editor)
                self.closeEditor(editor, QAbstractItemDelegate.NoHint)
        else:
            s = self.selection()
            if len(s) == 1:
                self.model().item_chosen(s[0])
            elif len(s) > 0 and self.model().can_add_to_queue(s):
                self.model().add_to_queue(s)

    def details(self):
        return self.model().details(self.selection())

    def can_add_to_queue(self):
        return self.model().can_add_to_queue(self.selection())

    def can_remove(self):
        return self.model().can_remove(self.selection())

    def can_rename(self):
        return self.model().can_rename(self.selection())

    def can_set_priority(self, prio):
        return self.model().can_set_priority(self.selection(), prio)

    def add_selected_to_queue(self, play=False, replace=False):
        self.model().add_to_queue(self.selection(), play, replace)

    def remove_selected(self):
        self.model().remove(self.selection())

    def rename_selected(self):
        s = self.selection()
        if len(s) == 1 and self.can_rename():
            i = self.model().index(s[0], 0)
            self.setCurrentIndex(i)
            self.edit(i)

    def set_priority(self, prio):
        self.model().set_priority(self.selection(), prio)

class PathBar(QWidget):

    clicked = Signal(str)

    def __init__(self, parent):
        super().__init__(parent)
        self.path = []
        self._mapper = QSignalMapper(self)
        self._mapper.mapped.connect(self._clicked)
        self._label = QLabel(self)
        self._label.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred))
        self._layout = QHBoxLayout(self)
        self._layout.addWidget(self._label)
        self._layout.setContentsMargins(0, 4, 0, 0)
        self.buttons = []
        self.actions = []
        self.set_path('')

    def _make_button(self):
        number = len(self.buttons)
        a = QAction(self)
        a.triggered.connect(self._mapper.map)
        self._mapper.setMapping(a, number)
        b = QToolButton(self)
        b.setDefaultAction(a)
        b.setSizePolicy(QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred))
        self._layout.insertWidget(number, b)
        self.actions.append(a)
        self.buttons.append(b)

    def set_root_icon(self, theme_name):
        self.actions[0].setIcon(QIcon.fromTheme(theme_name))

    def _clicked(self, number):
        self.clicked.emit('/'.join(self.path[:number]))

    def set_path(self, path):
        self.path = path.strip('/').split('/')
        if self.path == ['']:
            self.path = []
        for i in range(len(self.buttons), len(self.path) + 1):
            self._make_button()
        for i in range(1, len(self.path) + 1):
            self.actions[i].setText(self.path[i-1])
            self.buttons[i].show()
        for i in range(len(self.path) + 1, len(self.buttons)):
            self.buttons[i].hide()





import os
import datetime
import subprocess

version_info = (1, 0, 0, 'beta', 3)

def _get_version():
    v = '%s.%s' % version_info[:2]
    if version_info[2] != 0:
        v += '.%s' % version_info[2]
    v += '%s%s' % ({
        'alpha': 'a',
        'beta': 'b',
        'rc': 'c',
        'final': '',
    }[version_info[3]], version_info[4])

    repo_dir = os.path.dirname(os.path.abspath(__file__))
    timestamp_file = os.path.join(repo_dir, 'gittimestamp.txt')

    if version_info[3] == 'alpha':
        git_log = subprocess.Popen('git log --pretty=format:%ct --quiet -1 HEAD',
                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                shell=True, cwd=repo_dir, universal_newlines=True)
        timestamp = git_log.communicate()[0]
        try:
            timestamp = datetime.datetime.utcfromtimestamp(int(timestamp))
            timestamp = timestamp.strftime('%Y%m%d%H%M%S')
        except ValueError:
            timestamp = '0'
        if timestamp != '0':
            with open(timestamp_file, 'w') as f:
                f.write(timestamp + '\n')
        elif os.path.isfile(timestamp_file):
            with open(timestamp_file, 'r') as f:
                timestamp = f.read().strip()
        v += '.dev' + timestamp

    else:
        try:
            os.remove(timestamp_file)
        except:
            pass

    return v

version = _get_version()



#!/usr/bin/env python3

import os
import glob
import shutil
from distutils.core import setup
from distutils.command.build import build
from distutils.command.clean import clean
from distutils.command.sdist import sdist
from distutils.spawn import spawn

from qygmy.__version__ import version, version_info
SRC_DIR = os.path.dirname(os.path.abspath(__file__))


class build(build):
    def run(self):
        os.chdir(SRC_DIR)
        spawn(['make'])
        super().run()

class clean(clean):
    def run(self):
        os.chdir(SRC_DIR)
        spawn(['make', 'clean'])
        super().run()

class deb(sdist):
    def run(self):
        spawn(['make', 'dchrelease'])
        self.dist_dir = os.path.join(self.dist_dir, 'deb')
        super().run()
        name = self.distribution.get_name()
        version = self.distribution.get_version()
        ext = 'tar.gz'
        sourcedir = name + '-' + version
        tar = sourcedir + '.' + ext
        debversion = version.replace('.dev', '~dev')
        origtar = name + '_' + debversion + '.orig.' + ext
        os.chdir(self.dist_dir)
        spawn(['tar', 'xvzf', tar])
        os.rename(tar, origtar)
        os.chdir(sourcedir)
        spawn(['debuild'])
        os.chdir('..')
        shutil.rmtree(sourcedir, True)
        for i in glob.glob(name + '_' + debversion + '-*.build'):
            os.remove(i)


setup(
    name = 'qygmy',
    version = version,
    description = "Simple graphical MPD client.",
    long_description = """Simple graphical MPD client written in Python and Qt/PySide. It's interface
is strongly inspired by Pygmy (the GTK client).""",
    author = 'tsufeki',
    author_email = 'tsufeki@ymail.com',
    url = 'https://github.com/tsufeki/qygmy',
    download_url = 'https://github.com/tsufeki/qygmy/archive/v{}.tar.gz'.format(version),
    license = 'BSD',
    install_requires = [
        'PySide>=1.1.2',
        'python-mpd2',
    ],
    packages = [
        'qygmy',
        'qygmy.templates',
        'qygmy.ui',
    ],
    package_data = {
        'qygmy': ['translations/*.qm', 'gittimestamp.txt'],
    },
    scripts = ['bin/qygmyrun'],
    data_files = [
        ('share/applications', ['qygmy.desktop']),
        ('share/man/man1', ['qygmy.1']),
    ],
    classifiers = [
        {
            'alpha': 'Development Status :: 3 - Alpha',
            'beta':  'Development Status :: 4 - Beta',
            'rc':    'Development Status :: 4 - Beta',
            'final': 'Development Status :: 5 - Production/Stable',
        }[version_info[3]],
        'Environment :: X11 Applications :: Qt',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Topic :: Multimedia :: Sound/Audio',
    ],
    cmdclass = {
        'clean': clean,
        'build': build,
        'deb': deb,
    },
)
